#if defined(TROOPER)

#ifdef VXWORKS
#include <vxWorks.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <math.h>

#include "libp3d.h"
#include "p3dInt.h"
#include "p3dFullImpl.h"

/*
 * All This Stuff Is Called OnLy for rTrooper Model
 */

static bool _trajectory_is_safe(P3DPlanner *planner, P3D_MAP *map, double x,
		double y, double theta, double r, P3DRobotConfig *start);
		
static double p3dArcGetMaxSpeed(const P3DPlanner *planner, const P3DArc *arcs, int n, 
		double v0, double min, double max);

bool
trajectory_is_safe(P3DPlanner *planner, P3D_MAP *map, 
			double robot_x, double robot_y, double robot_theta, double alpha)
{
	double curv = alpha / p3dGetDistanceFixAndMobileAxle(planner);
	bool res;
	T3D mainToCenterOfMap;
	T3D* mainToOrigin;
	int quality;
	P3DRobotConfig start;

	mainToOrigin = t3dNew(T3D_EULER, T3D_ALLOW_CONVERSION);
	mainToOrigin->euler.euler[0] = robot_theta;
	mainToOrigin->euler.euler[3] = robot_x;
	mainToOrigin->euler.euler[4] = robot_y;

	p3dComputeRobotPositionInMap(map, mainToOrigin, &mainToCenterOfMap,
			NULL, NULL);

	/* Compute the initial configuration  */
	quality = PLACE_ROBOT(planner->robot, planner->geom, 
			mainToCenterOfMap.euler.euler[3],
			mainToCenterOfMap.euler.euler[4],
			mainToCenterOfMap.euler.euler[0],
			map, &start);

    fprintf(stderr, "executing trajectory_is_safe (x=%lf, y=%lf, theta=%lf)... \n", 
			mainToCenterOfMap.euler.euler[3],
			mainToCenterOfMap.euler.euler[4],
			mainToCenterOfMap.euler.euler[0]);

	res = _trajectory_is_safe(planner, map,
			mainToCenterOfMap.euler.euler[3],
			mainToCenterOfMap.euler.euler[4],
			mainToCenterOfMap.euler.euler[0],
			curv, &start);

	t3dDelete(mainToOrigin);
	return res;
}

int
p3dAco(P3DPlanner *planner, P3D_MAP *map, 
		T3D* mainToOrigin, 
		P3DPlanSolution * sol,
		double alpha, double v, double max_angle, double min_angle)
{
	double L = p3dGetDistanceFixAndMobileAxle(planner);
	double curv = alpha / L;
	double min_curv = min_angle / L;
	double max_curv = max_angle / L;
	double x, y, theta;
	T3D mainToCenterOfMap;
	struct timeval begin, end;
	double diff;
	double stop_distance;
	P3DArc *arcs_alpha, *arc_alpha;	
	double v_max, v_max_arcs, v_max_alpha;
	int ctrl = 1;

	gettimeofday(&begin, NULL);

	p3dComputeRobotPositionInMap(map, mainToOrigin, &mainToCenterOfMap,
			NULL, NULL);

   /* 
	* Compute the position that the robot will approximatvly have
	* when p3dAco will be finished. For that, use the average time
	* and the previous speed orders
	*/

   x = mainToCenterOfMap.euler.euler[3];
   y = mainToCenterOfMap.euler.euler[4];
   theta = mainToCenterOfMap.euler.euler[0];
   
   p3dComputeRobotMovement(&x, &y, &theta, sol, planner);

   stop_distance = MAX(p3dGetStopDistance(v, planner), planner->dMin);

   /* 
	* Recompute the node position with the new stop_distance 
	* it may be better to only recompute if stop_distance_previous != stop_distance
	* */
   p3dArcComputeNodePosition(planner->arcs, planner->nArcs, 0.0, 0.0, 0.0, stop_distance);

   /* Compute the special arc for alpha */
   /* We can't call p3dArcsCreate for 1 arcs because there is a nice bug. nR
	* will become 0 and leads to a->r = nan*/
   /* Don't forgive it creates 3 Arcs but we are only interested in the last one */
   arcs_alpha = p3dArcsCreate(2, planner->arcs->n, stop_distance, 0.0, 0.0, 0.0,
		   1/curv, 1);

   arc_alpha = (arcs_alpha+2);

   /* Compute the cost of each arcs */
   p3dArcComputeCostLimited(planner, map, planner->arcs, planner->nArcs, 
		   x, y, theta, 0.0, min_curv, max_curv);
   p3dArcComputeCostLimited(planner, map, arc_alpha, 1, x, y, theta, 0.0, min_curv, max_curv);

   /* Compute the min speed possible */
   v_max_arcs = p3dArcGetMaxSpeed(planner, planner->arcs, planner->nArcs, sol->v, min_curv, max_curv);
   v_max_alpha = p3dArcGetMaxSpeed(planner, arc_alpha, 1, sol->v, min_curv, max_curv);
   
   if (v_max_arcs < v_max_alpha) 
	   v_max = v_max_arcs;
   else 
	   v_max = v_max_alpha;

   /* The trajectory at v is ok */
   if (v_max > v) {
	   ctrl = 0;
	   v_max = v;
   }

   p3dArcsDestroy(arcs_alpha, 3);

   sol->v = v_max;
   sol->alpha = alpha;

   gettimeofday(&end, NULL);

   diff = sub_timeval(&begin, &end);
   sol->average_time = ( sol->average_time * sol->ncalls + diff ) / (sol->ncalls+1);
   sol->ncalls++;

   return ctrl;
}

static bool
_trajectory_is_safe(P3DPlanner *planner, P3D_MAP *map, double x, double y,
		double theta, double r, P3DRobotConfig *start)
		
{
	// XXX Of course, we must do something smarter
	double ds = 0.2;
	double nb_nodes = 40;
	P3DRobotConfig config1, config2;
	P3DRobotConfig *curr = &config1;
	P3DRobotConfig *prev = start;
	int i;

	for (i = 1; i <= nb_nodes; ++i) 
	{
		double x_new, y_new, theta_new;
		double quality, cost;
		p3dGetArcPos(r, ds*i, x, y, theta, &x_new, &y_new, &theta_new);

		quality = PLACE_ROBOT(planner->robot, planner->geom, x_new, y_new, theta_new, 
				map, curr);

		/* For moment we consider that a trajectory is impossible only if
		 * quality is < 0
		 * The trajectory may be hard to follow but usable. 
		 */

		if (quality < 0.0) {
			fprintf(stderr, "Placement failed for node %d : (%lf,%lf,%lf)\n",
					i, x_new, y_new, theta_new);
			fprintf(stderr, "Config quality %lf\n", curr->q);
			return false;
		}
		
		/*
		 * Compute the cost of the position, particulary if the delta with the
		 * previous position is not too important
		 */
		cost = EVAL_CONFIG(planner->geom, curr, prev);
		if (cost < 0.0) {
			fprintf(stderr, "Placement impossible for node %d : (%lf,%lf,%lf)\n",
					i, x_new, y_new, theta_new);
			fprintf(stderr, "Config cost %lf\n", cost);
			return false;
		}

		/* Switch prev and curr */
		if (curr == &config1) {
			curr = &config2;
			prev = &config1;
		} else {
			curr = &config1;
			prev = &config2;
		}
	}
	return true;
}

static double 
p3dArcGetMaxSpeed(const P3DPlanner *planner, const P3DArc *arcs, int n, double v0,
		double min, double max)
{
	const P3DArc *a;
	int i;
	double speed_max = planner->robot->vMax;

	for (i = 0, a = arcs; i < n && speed_max > 0; i++, a++) {
		if (a->r < min || a->r > max) 
			continue;

		double speed = max_speed(a, planner);

		if (speed < v0) { /* it means that we can enter in an obstacle or in
						   unknow without the capaticy of stopping */ 
			speed_max = -1.0;
		}

		if (speed < speed_max) 
			speed_max = speed;
	}

	return speed_max;
}

#endif
