/*
 * Copyright (C) 2001 LAAS/CNRS 
 *
	     Written by T. Simeon 
	     july 91

  This package provides a set of functions for manipulating AVL trees.

  These trees are binary and equilibrated search trees for which 
  the insertion, deletion or search operations can be performed 
  in O(log n).
  
  The main routines are:

     (int)AVLInsertNode((void **)avl , (void *)key , fct_compar)
     (int)AVLDeleteNode((void **)avl , (void *)key , fct_compar)
  (void *)AVLFindNode  ((void *) avl , (void *)key , fct_compar)
  
  They are generalized from Knuth (6.2.3). All comparisons are done with
  a user-supplied routine. This routine is called with two arguments.
  The first one is a pointer to the new key, the other one is a pointer 
  to the current key. The routine must return -1, 0 or 1 according to
  whether the first argument is to be considered less than, equal to or
  greater than the second argument. The comparaison function need not
  compare every byte, so arbitrary data may be contained in the elements
  in addtion to the value being compared.

  In addition, the following routines are available:

  (int)AVLCheckBalance((void *)avl)
       returns TRUE if the tree is well balanced (only for debugging purpose)

  (int)AVLFindDeleteFirstNode((void *)avl)				      
       returns the best node of the tree

  (int)AVLHeight((void *)avl)
  (int)AVLWidth((void *)avl)
  (int)AVLLength((void *)avl)
      return the height,the number of leaves and the total number of
      nodes in the tree.

 *
 * $Log: avl.c,v $
 * Revision 1.1  2006/10/17 18:09:42  sjoyeux
 * Initial revision
 *
 * Revision 1.2  2001/07/12 17:38:45  mallet
 * sync
 *
 * Revision 1.1  2001/07/11 18:05:40  mallet
 * re-ecriture de p3d
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "avl.h"

#define MAX(x,y) ((x)>(y)?(x):(y))

static AVLNode *	makeNode(void *key);
static void		freeNode(AVLNode *n);

static int		rotateLeft(AVLNode **p_na);
static int		rotateRight(AVLNode **p_na);
static int		rotate(AVLNode **p_avl, short bal);

static int		changeLeftSubtree(AVLNode **p_avl, void **key);

static int		deleteThisNode(AVLNode **p_avl);
static int		deleteNode(AVLNode **p_avl, void *key, int
				(*fct_order)(void *, void *), int *res);

static int		findDeleteFirstNode(AVLNode **p_avl, void **key);
static int		findDeleteLastNode(AVLNode **p_avl, void **key);


/*
 * makeNode -------------------------------------------------------------
 *
 * Return storage for a new node. For efficiency, a pool of free nodes is
 * maintained, using the "left" field of AVLNode to chain the list.
 * Storage must be released with freeNode.
 */

static AVLNode *pool = NULL;

static AVLNode *
makeNode(void *key)
{
   AVLNode *n;

   if (pool != NULL) {
      n = pool;
      pool = pool->left;
   } else {
      n = malloc(sizeof(*n));
   }

   if (n == NULL) {
      fprintf(stderr, "p3d: can't create a new Avl node\n");
      return(NULL);
   }

   n->key    = key;
   n->left   = NULL;
   n->right  = NULL;
   n->bal    = 0;

   return n;
}

static void
freeNode(AVLNode *n)
{
   if (n == NULL) return;

   n->left = pool;
   pool = n;
}


/*
 * rotateLeft -----------------------------------------------------------
 *
 */

static int
rotateLeft(AVLNode **p_na)
{
   register AVLNode *na = *p_na;
   register AVLNode *nb = AVL_GET_LEFT(na);
   short    bb = AVL_GET_BAL(nb);
   AVLNode *nc;
   short    bc;

   return 
    (AVL_GET_BAL(nb) == L_BAL) ?
      (AVL_SET_BAL(na,E_BAL),
       AVL_SET_BAL(nb,E_BAL),
       AVL_SET_LEFT (na , AVL_GET_RIGHT(nb)),
       AVL_SET_RIGHT(nb , na),
       *p_na = nb,
       TRUE):

	 (bb == E_BAL) ?
	   (AVL_SET_BAL(na,L_BAL),
	    AVL_SET_BAL(nb,R_BAL),
	    AVL_SET_LEFT (na , AVL_GET_RIGHT(nb)),
	    AVL_SET_RIGHT(nb , na),
	    *p_na = nb,
	    FALSE):
	      
	      (nc = AVL_GET_RIGHT(nb),
	       bc = AVL_GET_BAL(nc),
	       AVL_SET_BAL(nc,E_BAL),
	       (bc == L_BAL) ? 
	       (AVL_SET_BAL(na,R_BAL),
		AVL_SET_BAL(nb,E_BAL)):
	       (bc == E_BAL) ?
	       (AVL_SET_BAL(na,E_BAL),
		AVL_SET_BAL(nb,E_BAL)):
	       (AVL_SET_BAL(na,E_BAL),
		AVL_SET_BAL(nb,L_BAL)),
	       AVL_SET_RIGHT(nb,AVL_GET_LEFT(nc)),
	       AVL_SET_LEFT(nc,nb),
	       AVL_SET_LEFT(na,AVL_GET_RIGHT(nc)),
	       AVL_SET_RIGHT(nc,na),
	       *p_na = nc,
	       TRUE);
}


/*
 * rotateRight ----------------------------------------------------------
 *
 */

static int
rotateRight(AVLNode **p_na)
{
   register AVLNode *na = *p_na;
   register AVLNode *nb = AVL_GET_RIGHT(na);
   short    bb = AVL_GET_BAL(nb);
   AVLNode *nc;
   short    bc;

   return 
    (AVL_GET_BAL(nb) == R_BAL) ?
      (AVL_SET_BAL(na,E_BAL),
       AVL_SET_BAL(nb,E_BAL),
       AVL_SET_RIGHT(na , AVL_GET_LEFT(nb)),
       AVL_SET_LEFT(nb , na),
       *p_na = nb,
       TRUE):

	 (bb == E_BAL) ?
	   (AVL_SET_BAL(na,R_BAL),
	    AVL_SET_BAL(nb,L_BAL),
	    AVL_SET_RIGHT(na , AVL_GET_LEFT(nb)),
	    AVL_SET_LEFT(nb , na),
	    *p_na = nb,
	    FALSE):
	      
	      (nc = AVL_GET_LEFT(nb),
	       bc = AVL_GET_BAL(nc),
	       AVL_SET_BAL(nc,E_BAL),
	       (bc == L_BAL) ? 
	       (AVL_SET_BAL(na,E_BAL),
		AVL_SET_BAL(nb,R_BAL)):
	       (bc == E_BAL) ?
	       (AVL_SET_BAL(na,E_BAL),
		AVL_SET_BAL(nb,E_BAL)):
	       (AVL_SET_BAL(na,L_BAL),
		AVL_SET_BAL(nb,E_BAL)),
	       AVL_SET_LEFT(nb,AVL_GET_RIGHT(nc)),
	       AVL_SET_RIGHT(nc,nb),
	       AVL_SET_RIGHT(na,AVL_GET_LEFT(nc)),
	       AVL_SET_LEFT(nc,na),
	       *p_na = nc,
	       TRUE);
}


/*
 * rotate ---------------------------------------------------------------
 *
 */
    
static int
rotate(AVLNode **p_avl, short bal)
{
   return (bal < L_BAL) ? rotateLeft(p_avl):
      (bal > R_BAL) ? rotateRight(p_avl):
      !bal;
}


/*
 * changeLeftSubtree ----------------------------------------------------
 *
 */

static int
changeLeftSubtree(AVLNode **p_avl, void **key)
{
   AVLNode *node       = *p_avl;
   AVLNode *left_son   = AVL_GET_LEFT(node);
   AVLNode **right_son = &AVL_GET_RIGHT(node);
   short    newbal;

   if (*right_son) {

      if(changeLeftSubtree(right_son,key)) {

	 if((newbal = AVL_ADD_BAL(node,L_BAL)) < L_BAL)
	    return(rotateLeft(p_avl));
	 else
	    return(!newbal);

      } else
	 return(HEIGHT_NO_MODIFIED);

   } else {
      *key = AVL_GET_KEY(node);

      if(left_son) {
	 AVL_SET_KEY(node,AVL_GET_KEY(left_son));
	 AVL_DEL_LEFT(node);
      } else {
	 freeNode(node);
	 *p_avl = NULL;
      }
      return(HEIGHT_MODIFIED);
   }
}


/*
 * deleteThisNode -------------------------------------------------------
 *
 */

static int
deleteThisNode(AVLNode **p_avl)
{
   AVLNode *node       = *p_avl;
   AVLNode **left_son  = &AVL_GET_LEFT(node);
   AVLNode *right_son  = AVL_GET_RIGHT(node);
   short    newbal;

   if(*left_son) {
      if(changeLeftSubtree(left_son, &AVL_GET_KEY(node))) {

	 if((newbal = AVL_ADD_BAL(node,R_BAL)) > R_BAL)
	    return(rotateRight(p_avl));
	 else 
	    return(!newbal);
      } else
	 return(HEIGHT_NO_MODIFIED);
   } else {
      if(right_son) {
	 AVL_SET_KEY(node,AVL_GET_KEY(right_son)); 
	 AVL_DEL_RIGHT(node); 
      } else {
	 freeNode(node);
	 *p_avl = NULL;
      }

      return(HEIGHT_MODIFIED);
   }
}  


/*
 * insertNode -----------------------------------------------------------
 *
 */

static int
insertNode(AVLNode **p_avl, void *key,
	   int (*fct_order)(void *, void *), int *res)
{
   AVLNode *node = *p_avl;
   AVLNode **son;
   short    sens;
   short    newbal;
	     
   if ((sens = (*fct_order)(key,AVL_GET_KEY(node)))) {
      son = AVL_GET_ADRSON(node,sens);
  
      if(*son) {
	 if(insertNode(son,key,fct_order,res)) {
	    newbal = AVL_ADD_BAL(node,sens);

	    if(AVL_BAL_EQUILIBRATED(newbal)) 
	       return(abs(newbal));
	    else {
	       rotate(p_avl,newbal);
	       return(HEIGHT_NO_MODIFIED);
	    }
	 } else
	    return(HEIGHT_NO_MODIFIED);
      } else {
	 AVL_ADD_SON(node,sens,key);

	 if(!AVL_GET_SON(node,sens)) {
	    *res = FALSE;
	    return(HEIGHT_NO_MODIFIED);
	 }
	
	 if(AVL_GET_OTHERSON(node,sens))
	    return(HEIGHT_NO_MODIFIED);
	 else
	    return(HEIGHT_MODIFIED);
      }
   } else {
      *res = FALSE;
      return(HEIGHT_NO_MODIFIED);
   }
}


/*
 * AVLInsertNode --------------------------------------------------------
 *
 */

int
AVLInsertNode(AVLNode **p_avl, void *key, int (*fct_order)(void *, void *))
{
   if(*p_avl) {
      int res = TRUE;
      insertNode(p_avl,key,fct_order,&res);
      return(res);
   } else {
      *p_avl = makeNode(key);
      return(TRUE);
   }
}


/*
 * AVLFindNode ----------------------------------------------------------
 *
 */

void *
AVLFindNode(AVLNode *avl, void *key, int (*fct_order)(void *, void *))
{
   void  *keyavl;
   short sens;

   if(!avl) return(NULL);

   keyavl = AVL_GET_KEY(avl);
  
   if((sens = (*fct_order)(key, keyavl)))
      return(AVLFindNode(AVL_GET_SON(avl,sens), key, fct_order));
   else
      return(keyavl);
}


/*
 * deleteNode -----------------------------------------------------------
 *
 */

static int
deleteNode(AVLNode **p_avl, void *key, int (*fct_order)(void *, void *),
	   int *res)
{
   AVLNode *node = *p_avl;
   AVLNode **son;
   short    sens;

   if ((sens = (*fct_order)(key, AVL_GET_KEY(node)))) {
      son = AVL_GET_ADRSON(node,sens);
    
      if(*son) {
	 if(deleteNode(son, key, fct_order, res)) 
	    return(rotate(p_avl,AVL_ADD_BAL(node,-sens)));
	 else
	    return(HEIGHT_NO_MODIFIED);
      } else {
	 *res = FALSE;
	 return(HEIGHT_NO_MODIFIED);
      }
   } else
      return(deleteThisNode(p_avl));
}


/*
 * AVLDeleteNode --------------------------------------------------------
 *
 */

int
AVLDeleteNode(AVLNode **p_avl, void *key, int (*fct_order)(void *, void *))
{
   if(*p_avl) {
      int res = TRUE;
      deleteNode(p_avl,key,fct_order,&res);
      return(res);
   } else
      return(FALSE);
}


/*
 * findDeleteFirstNode --------------------------------------------------
 *
 */
 
static int
findDeleteFirstNode(AVLNode **p_avl, void **key)
{
   AVLNode *node  = *p_avl;
   AVLNode **left = &AVL_GET_LEFT(node);
   AVLNode *right;

   if(*left) {

      if(findDeleteFirstNode(left,key)) 
	 return(rotate(p_avl,AVL_ADD_BAL(node,R_BAL)));
      else
	 return(HEIGHT_NO_MODIFIED);

   } else {

      *key = AVL_GET_KEY(node);
    
      if((right = AVL_GET_RIGHT(node))) {
	 AVL_SET_KEY(node,AVL_GET_KEY(right));
	 AVL_DEL_RIGHT(node);
      } else {
	 freeNode(node);
	 *p_avl = NULL;
      }

      return(HEIGHT_MODIFIED);
   }
}


/*
 * AVLFindDeleteFirstNode -----------------------------------------------
 *
 */

void *
AVLFindDeleteFirstNode(AVLNode **p_avl)
{
   if(*p_avl) {
      void *res;
      findDeleteFirstNode(p_avl,&res);
      return(res);
   } else
      return(NULL);
}


/*
 * findDeleteLastNode ---------------------------------------------------
 *
 */

static int
findDeleteLastNode(AVLNode **p_avl, void **key)
{
   AVLNode *node   = *p_avl;
   AVLNode **right = &AVL_GET_RIGHT(node);
   AVLNode *left;

   if(*right) {

      if(findDeleteLastNode(right,key)) 
	 return(rotate(p_avl,AVL_ADD_BAL(node,L_BAL)));
      else
	 return(HEIGHT_NO_MODIFIED);

   } else {

      *key = AVL_GET_KEY(node);
    
      if((left = AVL_GET_LEFT(node))) {
	 AVL_SET_KEY(node,AVL_GET_KEY(left));
	 AVL_DEL_LEFT(node);
      } else {
	 freeNode(node);
	 *p_avl = NULL;
      }

      return(HEIGHT_MODIFIED);
   }
}


/*
 * AVLFindDeleteLastNode ------------------------------------------------
 *
 */

void *
AVLFindDeleteLastNode(AVLNode **p_avl)
{
   if(*p_avl) {
      void *res;
      findDeleteLastNode(p_avl,&res);
      return(res);
   } else
      return(NULL);
}


/*
 * AVLCheckBalance ------------------------------------------------------
 *
 */

int
AVLCheckBalance(AVLNode *avl)
{
   int h_l = 0;
   int h_r = 0;

   if(!avl) return(0);

   if(AVL_GET_LEFT(avl))
      h_l = AVLCheckBalance(AVL_GET_LEFT(avl));

   if(AVL_GET_RIGHT(avl))
      h_r = AVLCheckBalance(AVL_GET_RIGHT(avl));

   if(h_r - h_l != avl->bal)
      fprintf(stderr,"p3d: (AVL) Warning: Error in the balance computation\n");

   if(!AVL_BAL_EQUILIBRATED(avl->bal))
      fprintf(stderr,"p3d: (AVL) Warning: AVL non well balanced\n");

   return(1 + MAX(h_l,h_r));
}


/*
 * AVLHeight ------------------------------------------------------------
 *
 */

int
AVLHeight(AVLNode *avl)
{
   if(!avl) return(-1);
   return(1 + MAX(AVLHeight(avl->left),AVLHeight(avl->right)));
}


/*
 * AVLWidth -------------------------------------------------------------
 *
 */

int
AVLWidth(AVLNode *avl)
{
   int w;

   if(!avl) return(0);

   w = AVLWidth(AVL_GET_LEFT(avl)) + AVLWidth(AVL_GET_RIGHT(avl));

   return(MAX(1,w));
}


/*
 * AVLLength ------------------------------------------------------------
 *
 */

int
AVLLength(AVLNode *avl)
{
   int w;

   if(!avl) return(0);

   w = 1 + AVLLength(AVL_GET_LEFT(avl)) + AVLLength(AVL_GET_RIGHT(avl));
   return(w);
}
